import pytest
from .validator import assert_valid_schema


@pytest.mark.organization
@pytest.mark.parametrize('sid', [
    ('SIBYLLA'),
    ('PROTECTORA'),
    ('ODINT'),
    ('HWFC'),
    ('STARD'),
])
@pytest.mark.organization
def test_request(sid):
    from rsi_scraper import Organization

    object = Organization(sid)
    data = object.execute()

    # with open("./test.json", 'w') as f:
    #     f.write(str(data))

    assert data is not None

    assert_valid_schema(data, 'organization.json')



from .connector import Connector
from .interface import ICommand
from .organization import Organization, OrganizationMembers
from .progress_tracker import ProgressTracker, ProgressTrackerInfo
from .roadmap import Roadmap
from .ship import Ship
from .starmap import (
    StarmapSystems, StarmapTunnels, StarmapSpecies,
    StarmapAffiliations, StarmapStarSystems,
    StarmapCelestialObjects, StarmapSearch,
    StarmapRouteSearch
)
from .stats import Stats
from .telemetry import Telemetry
from .user import User
from .version import Version

__all__ = [
    'Connector',
    'ICommand',
    'Organization', 'OrganizationMembers',
    'ProgressTracker', 'ProgressTrackerInfo',
    'Roadmap',
    'Ship',
    'StarmapSystems', 'StarmapTunnels', 'StarmapSpecies',
    'StarmapAffiliations', 'StarmapStarSystems',
    'StarmapCelestialObjects', 'StarmapSearch',
    'StarmapRouteSearch',
    'Stats',
    'Telemetry',
    'User',
    'Version',
]

__version__ = '0.6.8'



"""Provide methods to build components
"""


class ICommand:
    """Provide methods to build components.
    """

    def execute(self):
        """Execute the module.

        Raises:
            NotImplementedError:
        """
        raise NotImplementedError()

    async def execute_async(self):
        """Execute the module asynchronously.

        Raises:
            NotImplementedError:
        """
        raise NotImplementedError()

    @staticmethod
    def convert_val(val: object):
        """Convert an object to HTTP parameter.

        Args:
          val (object): A value to convert in str.

        Returns:
            str: A string ready for HTTP request.
        """
        if val is None:
            return ""
        if isinstance(val, list) and len(val) == 1:
            if isinstance(val, int):
                return int(val)
            return val[0]

        return val

    @staticmethod
    def http_formatter(val1, val2):
        """Concatenate two objects to create http friendly string.

        Args:
            val1 (object): First value.
            val2 (object): Second value.

        Returns:
            string: A string HTTP ready.
        """
        line = f"{val1},{val2}"
        if len(line) == 1:
            line = ''
        return line



import pytest
from .validator import assert_valid_schema


@pytest.mark.user
@pytest.mark.parametrize('handle', [
    ("dymerz"),
    ("Dewderonomy")
])
def test_request(handle):
    from rsi_scraper import User

    object = User(handle)
    data = object.execute()
    assert data is not None

    assert_valid_schema(data, 'user.json')



import pytest
from .validator import assert_valid_schema


@pytest.mark.telemetry
def test_request():
    from rsi_scraper import Telemetry

    object = Telemetry('MONTH', '3.7')
    data = object.execute()
    assert data is not None

    assert_valid_schema(data, 'telemetry.json')



"""Roadmap
"""
from .connector import Connector
from .interface import ICommand


class ProgressTracker(ICommand):
    """ProgressTracker
    """

    __url_progress_tracker = "https://robertsspaceindustries.com/graphql"

    def __init__(self, date_min="2020-01-01", date_max="2022-12-31"):
        """
        Args:
            date_min (str): The roadmap start date.
            date_max (str): The roadmap end date.
        """
        self.date_min = date_min
        self.date_max = date_max
        self.limit = 20
        self.offset = 0

    async def execute_async(self):
        return self.execute()

    def execute(self):
        return self.get_teams()

    def get_teams(self):
        data = [
            {
                "operationName": "teams",
                "query": """query
                    teams(
                        $startDate: String!,
                        $endDate: String!,
                        $search: String,
                        $teamSlug: String,
                        $deliverableSlug: String,
                        $projectSlugs: [String],
                        $disciplineSlugs: [String],
                        $sortBy: SortMethod,
                        $offset: Int,
                        $limit: Int
                    ) {
                        progressTracker {
                            teams(
                                startDate: $startDate
                                endDate: $endDate
                                search: $search
                                teamSlug: $teamSlug
                                deliverableSlug: $deliverableSlug
                                projectSlugs: $projectSlugs
                                disciplineSlugs: $disciplineSlugs
                                sortBy: $sortBy
                                offset: $offset
                                limit: $limit) {
                                    totalCount
                                    metaData {
                                        ...Team
                                        timeAllocations {
                                            ...TimeAllocation
                                            __typename
                                        }
                                        __typename
                                    }
                                    __typename
                                }
                            __typename
                        }
                    }
                    fragment Team on Team {
                        title
                        description
                        uuid
                        abbreviation
                        startDate
                        endDate
                        numberOfDeliverables
                        slug
                        __typename
                    }

                    fragment TimeAllocation on TimeAllocation {
                        startDate
                        endDate
                        uuid
                        partialTime
                        __typename
                    }
              """,
                "variables": {
                    "startDate": self.date_min,
                    "endDate": self.date_max,
                    "offset": self.offset,
                    "limit": self.limit
                }
            }
        ]
        req = Connector().request(self.__url_progress_tracker, method="post", json_data=data)
        if req is None or req.status_code != 200:
            return None
        try:
            resp = req.json()
        except Exception as e:
            print(e, flush=True)
            return None

        if len(resp) > 0 and 'data' not in resp[0]:
            return None

        return resp[0]['data']['progressTracker']['teams']['metaData']


class ProgressTrackerInfo(ICommand):
    """ProgressTrackerInfo
    """

    __url_progress_tracker = "https://robertsspaceindustries.com/graphql"

    def __init__(self, slug: str, date_min="2020-01-01", date_max="2022-12-31"):
        """
        Args:
            slug (str): The slug identifier to get.
            date_min (str): The roadmap start date.
            date_max (str): The roadmap end date.
        """
        self.team_slug = slug
        self.date_min = date_min
        self.date_max = date_max

    async def execute_async(self):
        return self.execute()

    def execute(self):
        deliverables = self.get_deliverables(self.team_slug)
        disciplines_slugs = [deliv['slug'] for deliv in deliverables]

        disciplines = self.get_disciplines(self.team_slug, disciplines_slugs)

        for i in range(len(deliverables)):
            deliverables[i]['disciplines'] = disciplines[i]
        return deliverables

    def get_deliverables(self, team_slug):
        data = [
            {
                "operationName": "deliverables",
                "query": """query
                    deliverables(
                        $startDate: String!,
                        $endDate: String!,
                        $search: String,
                        $deliverableSlug: String,
                        $teamSlug: String,
                        $projectSlugs: [String],
                        $categoryIds: [Int],
                        $sortBy: SortMethod,
                        $offset: Int,
                        $limit: Int
                    ) {
                        progressTracker {
                            deliverables(
                                startDate: $startDate
                                endDate: $endDate
                                search: $search
                                deliverableSlug: $deliverableSlug
                                teamSlug: $teamSlug
                                projectSlugs: $projectSlugs
                                categoryIds: $categoryIds
                                sortBy: $sortBy
                                offset: $offset
                                limit: $limit
                            ) {
                                totalCount
                                metaData {
                                    ...Deliverable card {
                                        ...Card
                                        __typename
                                    }
                                    projects {
                                        ...Project
                                        __typename
                                    }
                                    __typename
                                }
                                __typename
                            }
                            __typename
                        }
                    }
                    fragment Deliverable on Deliverable {
                        uuid
                        slug
                        title
                        description
                        startDate
                        endDate
                        numberOfDisciplines
                        numberOfTeams
                        updateDate
                        totalCount
                        __typename
                    }
                    fragment Card on Card {
                        id
                        title
                        description
                        category
                        release {
                            id
                            title
                            __typename
                        }
                        board {
                            id
                            title
                            __typename
                        }
                        updateDate
                        thumbnail
                        __typename
                    }
                    fragment Project on Project {
                        title
                        logo
                        __typename
                    }
                """,
                "variables": {
                    "teamSlug": team_slug,
                    "startDate": self.date_min,
                    "endDate": self.date_max,
                }
            }
        ]
        req = Connector().request(self.__url_progress_tracker, method="post", json_data=data)

        if req is None or req.status_code != 200:
            return None
        try:
            resp = req.json()
        except Exception as e:
            print(e, flush=True)
            return None

        if len(resp) > 0 and 'data' not in resp[0]:
            return None

        return resp[0]['data']['progressTracker']['deliverables']['metaData']

    def get_disciplines(self, team_slug: str, delivery_slugs: list()):
        data = []
        for del_slug in delivery_slugs:
            data.append({
                "operationName": "disciplines",
                "query": """query
                    disciplines($teamSlug: String! $deliverableSlug: String! $startDate: String! $endDate: String! ) {
                        progressTracker {
                            disciplines(teamSlug: $teamSlug deliverableSlug: $deliverableSlug startDate: $startDate endDate: $endDate ) {
                                ...Discipline
                                timeAllocations {
                                    ...TimeAllocation
                                    __typename
                                }
                                __typename
                            }
                            __typename
                        }
                    }
                    fragment Discipline on Discipline {
                        title
                        color
                        uuid
                        numberOfMembers
                        __typename
                    }
                    fragment TimeAllocation on TimeAllocation {
                        startDate
                        endDate
                        uuid
                        partialTime
                        __typename
                    }
                    """,
                "variables": {
                    "teamSlug": team_slug,
                    "deliverableSlug": del_slug,
                    "startDate": self.date_min,
                    "endDate": self.date_max,
                }
            })

        req = Connector().request(self.__url_progress_tracker, method="post", json_data=data)

        if req is None or req.status_code != 200:
            return None
        try:
            resp = req.json()
        except Exception as e:
            print(e, flush=True)
            return None

        if len(resp) > 0 and 'data' not in resp[0]:
            return None

        return [d['data']['progressTracker']['disciplines'][0] for d in resp]



"""
Organization
"""
import re
import queue

from threading import Thread
from lxml import html
from lxml import etree

from .connector import Connector
from .interface import ICommand


class Organization(ICommand):
    """Organization
    """

    __url_organization = "https://robertsspaceindustries.com/orgs/{0}"
    __url_search_orgs = "https://robertsspaceindustries.com/api/orgs/getOrgs"

    def __init__(self, sid: str):
        """
        Args:
            sid (str): The Spectrum Identification (sid).
        """
        self.sid = sid

    async def execute_async(self):
        return self.execute()

    def execute(self):
        que = queue.Queue()
        threads = []

        t0 = Thread(
            target=lambda q,
            arg1: q.put(self.get_organization_info(arg1)),
            args=(que, self.sid))
        t0.start()
        threads.append(t0)

        t1 = Thread(
            target=lambda q,
            arg1: q.put(self.search_orgs(arg1)),
            args=(que, self.sid))
        t1.start()
        threads.append(t1)

        for t in threads:
            t.join()

        counter = 0
        result = {}
        while not que.empty():
            res = que.get()

            if res is None:
                return None
            elif isinstance(res, list):
                if len(res) != 0:
                    res = res[0]
                else:
                    return None

            elif isinstance(res, dict):
                if not res:
                    return None

            result.update(res)
            counter += 1

        return result

    def get_organization_info(self, sid: str):
        """Get organization info.

        Args:
            sid (str): The Spectrum Identification (sid).

        Returns:
            dict: Organization information.
        """
        result = {}

        url = self.__url_organization.format(sid)

        # request website
        req = Connector().request(url, method="get")

        if req is None:
            return None
        if req.status_code == 404:
            return {}
        if req.status_code != 200:
            return None

        # get html contents
        tree = html.fromstring(req.content)

        result["url"] = url
        result["sid"] = sid

        for v in tree.xpath(
            '//*[contains(@class, "logo") and contains(@class, "noshadow")]/img/@src'
        ):
            result["logo"] = Connector.convert_path(v)
            break

        for v in tree.xpath('//*[@id="organization"]//h1/text()'):
            result["name"] = v.strip('/ ')
            break

        result["focus"] = {}
        result["focus"]["primary"] = {}

        for v in tree.xpath('//*[contains(@class, "primary") and contains(@class, "tooltip-wrap")]/img/@src'):
            result["focus"]["primary"]["image"] = Connector.convert_path(v)
            break
        for v in tree.xpath('//*[contains(@class, "primary") and contains(@class, "tooltip-wrap")]/img/@alt'):
            result["focus"]["primary"]["name"] = v.strip()
            break

        result["focus"]["secondary"] = {}
        for v in tree.xpath('//*[contains(@class, "secondary") and contains(@class, "tooltip-wrap")]/img/@src'):
            result["focus"]["secondary"]["image"] = Connector.convert_path(v)
            break
        for v in tree.xpath('//*[contains(@class, "secondary") and contains(@class, "tooltip-wrap")]/img/@alt'):
            result["focus"]["secondary"]["name"] = v.strip()
            break

        for v in tree.xpath('//*[contains(@class, "banner")]/img/@src'):
            result["banner"] = Connector.convert_path(v)
            break

        result["headline"] = {}
        for v in tree.xpath('//*[contains(@class, "body") and contains(@class, "markitup-text")]'):
            result["headline"]["html"] = etree.tostring(v, pretty_print=True).decode("utf-8")
            break

        for v in tree.xpath('//*[contains(@class, "body") and contains(@class, "markitup-text")]'):
            result["headline"]["plaintext"] = ''.join(v.itertext())
            break

        result["history"] = {}
        for v in tree.xpath('//*[@id="tab-history"]/div'):
            result["history"]["html"] = etree.tostring(v, pretty_print=True).decode("utf-8")
            break

        for v in tree.xpath('//*[@id="tab-history"]/div'):
            result["history"]["plaintext"] = ''.join(v.itertext())
            break

        result["manifesto"] = {}
        for v in tree.xpath('//*[@id="tab-manifesto"]/div'):
            result["manifesto"]["html"] = etree.tostring(v, pretty_print=True).decode("utf-8")
            break

        for v in tree.xpath('//*[@id="tab-manifesto"]/div'):
            result["manifesto"]["plaintext"] = ''.join(v.itertext())
            break

        result["charter"] = {}
        for v in tree.xpath('//*[@id="tab-charter"]/div'):
            result["charter"]["html"] = etree.tostring(v, pretty_print=True).decode("utf-8")
            break

        for v in tree.xpath('//*[@id="tab-charter"]/div'):
            result["charter"]["plaintext"] = ''.join(v.itertext())
            break

        return result

    def search_orgs(self, sid: str):
        """Search organizaion by SID ()

        Args:
            sid (str): Spectrum Identification of the organization

        Returns:
            dict: Organizations found
        """

        results = []
        page = 1

        payload = {
            "activity": [],
            "commitment": [],
            "language": [],
            "model": [],
            "pagesize": 12,
            "recruiting": [],
            "roleplay": [],
            "search": sid,
            "page": page,
            "size": [],
            "sort": ""
        }

        req = Connector().request(self.__url_search_orgs, payload)

        if req is None:
            return None
        elif req.status_code == 404:
            return {}
        elif req.status_code != 200:
            return None

        # get html contents
        result = req.json()

        if result["success"] == 0:
            print(result["code"] + ": " + result["msg"])
            return None

        row_html = result["data"]["html"]
        if row_html is None or row_html.strip() == "":
            return None

        # start process html
        tree = html.fromstring(row_html)

        for i in range(1, int(tree.xpath('count(//*[contains(@class, "org-cell")])')) + 1):
            org = {}
            try:

                for v in tree.xpath('//*[contains(@class, "org-cell")][{}]/a/*[@class="left"]/*[@class="identity"]/*[@class="symbol"]/text()'.format(i)):
                    org["sid"] = v.strip()
                    break

                # ensure the right Organization is returned
                if org["sid"] != sid:
                    continue

                for v in tree.xpath('//*[contains(@class, "org-cell")][{}]/a/@href'.format(i)):
                    org["href"] = Connector.convert_path(v)
                    break

                for v in tree.xpath('//*[contains(@class, "org-cell")][{}]/a/*[@class="left"]/*[@class="thumb"]/img/@src'.format(i)):
                    org["logo"] = Connector.convert_path(v)
                    break

                for v in tree.xpath('//*[contains(@class, "org-cell")][{}]/a/*[@class="left"]/*[@class="identity"]/*[contains(@class, "name")]/text()'.format(i)):
                    org["name"] = v
                    break

                for v in tree.xpath('//*[contains(@class, "org-cell")][{}]/a/*[@class="right"]/*[@class="infocontainer"][1]/*[@class="infoitem"][1]/*[@class="value"]/text()'.format(i)):
                    org["archetype"] = v
                    break

                for v in tree.xpath('//*[contains(@class, "org-cell")][{}]/a/*[@class="right"]/*[@class="infocontainer"][1]/*[@class="infoitem"][2]/*[contains(@class, "value")]/text()'.format(i)):
                    org["lang"] = v
                    break

                for v in tree.xpath('//*[contains(@class, "org-cell")][{}]/a/*[@class="right"]/*[@class="infocontainer"][1]/*[@class="infoitem"][3]/*[contains(@class, "value")]/text()'.format(i)):
                    org["commitment"] = v
                    break

                for v in tree.xpath('//*[contains(@class, "org-cell")][{}]/a/*[@class="right"]/*[@class="infocontainer"][2]/*[@class="infoitem"][1]/*[contains(@class, "value")]/text()'.format(i)):
                    org["recruiting"] = (v == "Yes")
                    break

                for v in tree.xpath('//*[contains(@class, "org-cell")][{}]/a/*[@class="right"]/*[@class="infocontainer"][2]/*[@class="infoitem"][2]/*[contains(@class, "value")]/text()'.format(i)):
                    org["roleplay"] = (v == "Yes")
                    break

                for v in tree.xpath('//*[contains(@class, "org-cell")][{}]/a/*[@class="right"]/*[@class="infocontainer"][2]/*[@class="infoitem"][3]/*[contains(@class, "value")]/text()'.format(i)):
                    org["members"] = int(v)
                    break

                results.append(org)
            except Exception as e:
                print(e, flush=True)
                continue

        return results


class OrganizationMembers(ICommand):

    __url_organization_members = "https://robertsspaceindustries.com/api/orgs/getOrgMembers"

    def __init__(self, sid: str, **kwargs):
        """
        Args:
            sid (str): The Spectrum Identification (sid).

        Keyword Args:
            rank (str): The id of the rank to search.
            role (str): The id of the role to search.
            main_org (bool): If true, get all the players which is the main organization.
        """
        self.sid = sid
        self.kwargs = kwargs

    async def execute_async(self):
        return self.execute()

    def execute(self):
        result = []
        page = self.convert_val(self.kwargs.get("page"))

        json_data = {
            "symbol": self.sid,
            "search": "",
            "pagesize": 32,
            "page": page
        }

        if 'rank' in self.kwargs:
            json_data['rank'] = self.convert_val(self.kwargs.get("rank"))
        elif 'role' in self.kwargs:
            json_data['role'] = self.convert_val(self.kwargs.get("role"))
        elif 'main_org' in self.kwargs:
            json_data['main_org'] = 1 if (self.convert_val(self.kwargs.get("main_org")).lower() == 'true') else 0

        # request website
        req = Connector().request(self.__url_organization_members, json_data=json_data, method="post")

        if req is None:
            return result
        if req.status_code != 200:
            return result
        res = req.json()

        if res['success'] == 0:
            if res['code'] == 'ErrApiThrottled':
                print("Failed, retrying", flush=True)
            else:
                print(f"Failed ({res['msg']})", flush=True)
            return result

        if res['data']['html'] == '':
            return result

        # get html contents
        tree = html.fromstring(res["data"]["html"])
        index = 1
        for _ in tree.xpath("//*[contains(@class, 'member-item')]"):
            user = {}

            for v in tree.xpath(f"//*[contains(@class, 'member-item')][{index}]//*[contains(@class, 'nick')]/text()"):
                user["handle"] = v.strip()
                break

            # check if the handle is already in result
            if user["handle"] in result:
                continue

            for v in tree.xpath(f"//*[contains(@class, 'member-item')][{index}]//*[contains(@class, ' name')]/text()"):
                user["display"] = v.strip()
                break

            for v in tree.xpath(f"//*[contains(@class, 'member-item')][{index}]//*[contains(@class, 'stars') and contains(@style, .)]"):
                style = v.attrib["style"]
                match = re.search(":\\s*([0-9]*)\\%", style, re.IGNORECASE)
                if match:
                    user["stars"] = int(int(match.group(1)) / 20)
                break

            for v in tree.xpath(f"//*[contains(@class, 'member-item')][{index}]//*[contains(@class, 'rank')]"):
                if v.attrib['class'] == 'rank':
                    user["rank"] = v.text.strip()
                    break

            user["roles"] = []
            for v in tree.xpath(f"//*[contains(@class, 'member-item')][{index}]//*[contains(@class, 'rolelist')]/li/text()"):
                user["roles"].append(v)

            for v in tree.xpath(f"//*[contains(@class, 'member-item')][{index}]//img/@src"):
                user["image"] = Connector.convert_path(v.strip())
                break

            # some fields are filled of "&nbsp", so skip them
            if user != {} and user["handle"] != "":
                result.append(user)
            index += 1

        return result



import pytest
from .validator import assert_valid_schema


@pytest.mark.roadmap
@pytest.mark.roadmap_starcitizen
def test_request_starcitizen():
    from rsi_scraper import Roadmap

    object = Roadmap('starcitizen', '3.12')
    data = object.execute()
    assert data is not None

    assert_valid_schema(data, 'roadmap_starcitizen.json')


@pytest.mark.roadmap
@pytest.mark.roadmap_squadron42
def test_request_squadron():
    from rsi_scraper import Roadmap

    object = Roadmap('squadron42')
    data = object.execute()
    assert data is not None

    assert_valid_schema(data, 'roadmap_squadron42.json')



import pytest
from .validator import assert_valid_schema


@pytest.mark.stats
def test_request():
    from rsi_scraper import Stats

    object = Stats()
    data = object.execute()
    assert data is not None

    assert_valid_schema(data, 'stats.json')



import os

import pytest
import warnings

from rsi_scraper import Connector
import requests

url = 'http://icanhazip.com'


@pytest.mark.connector
def test_request():
    conn = Connector()
    req = conn.request(url)
    assert req.status_code == 200


@pytest.mark.connector
async def test_request_async():
    conn = Connector()
    req = await conn.request_async(url)
    assert req.status_code == 200


@pytest.mark.connector
def test_proxy():

    req = requests.get(url, proxies={})
    ip_public = req.text.strip()
    assert req.status_code == 200

    conn = Connector()
    req = conn.request(url)
    ip_proxy = req.text.strip()
    assert req.status_code == 200

    if os.getenv('HTTP_PROXY'):
        if ip_public == ip_proxy:
            pytest.exit('Bad proxy configuration', 1)
    else:
        warnings.warn(UserWarning('Executing without proxy'))


@pytest.mark.connector
async def test_request_rsi():
    conn = Connector()
    req = await conn.request_async(conn.url_host)
    assert req.status_code == 200



import pytest
from .validator import assert_valid_schema


@pytest.mark.progress_tracker
def test_request():
    from rsi_scraper import ProgressTracker

    object = ProgressTracker()
    data = object.execute()
    assert data is not None

    assert_valid_schema(data, 'progress_tracker.json')


@pytest.mark.progress_tracker_info
def test_request_info():
    from rsi_scraper import ProgressTrackerInfo

    object = ProgressTrackerInfo('g42qg07ipfday')
    data = object.execute()
    assert data is not None

    assert_valid_schema(data, 'progress_tracker_info.json')



"""
A module to perform requests
"""

import os
import requests
from urllib.parse import urljoin


class Connector:

    url_host = 'https://robertsspaceindustries.com'
    __user_agent = f"""StarCitizen REST API/{
        os.getenv('VERSION') or 'DEVELOPMENT'
    } (starcitizen-api.com)"""

    async def request_async(
        self, url: str, json_data: dict = None,
        headers: dict = {}, method: str = "post"
    ):
        """An asynchronous request.

        Args:
            url (str): The URL.
            json_data (dict, optional): The data to send. Default to None.
            headers (dict, optional): Headers parameters. Default to {}.
            method (str, optional): Use the "get" or "post". Default to "post".

        Returns:
            requests.Response: The Response object
        """
        return self.request(url, json_data, headers, method)

    def request(
        self, url: str, json_data: dict = None,
        headers: dict = {}, method: str = "post"
    ):
        """Send a request to the specified url using parameters

        Args:
            url (str): The URL.
            json_data (dict, optional): The data to send. Default to None.
            headers (dict, optional): Headers parameters. Default to {}.
            method (str, optional): Use the "get" or "post". Default to "post".

        Returns:
            requests.Response: The Response object
        """
        headers['Accept-Language'] = 'en-US,en;q=0.5'
        headers['User-Agent'] = self.__user_agent
        headers['Cache-Control'] = "no-cache"
        headers['Cookie'] = "Rsi-Token="

        args = {
            "url": url,
            "headers": headers,
            "stream": False
        }

        if json_data is not None:
            args["json"] = json_data

        proxies = {}
        if os.getenv('HTTP_PROXY'):
            proxies = {'http': os.environ['HTTP_PROXY']}

        req = None
        if method == "post":
            req = requests.post(proxies=proxies, timeout=5, **args)
        elif method == "get":
            req = requests.get(proxies=proxies, timeout=5, **args)
        else:
            return None

        return req

    @classmethod
    def convert_path(cls, path: str):
        return urljoin(cls.url_host, path)



"""Telemetry
"""
import re
from lxml import html

from .connector import Connector
from .interface import ICommand


class Version(ICommand):
    """Version
    """

    __url_patch_notes = "https://robertsspaceindustries.com/patch-notes"
    __regex_website_version = r'((\d\.?)+)'
    __regex_website_background = r"(background-image\s*:\s*url\(')(.*)('\))"

    def __init__(self):
        pass

    async def execute_async(self):
        return self.execute()

    def execute(self):
        result = []
        req = Connector().request(self.__url_patch_notes, method="get")

        if req is None or req.status_code != 200:
            return None

        resp = req.text
        tree = html.fromstring(resp)
        for patch in tree.xpath("//*[contains(@class, 'content-block2') and contains(@class, 'hub-block')]"):
            p = {}

            # get article url
            href = patch.attrib['href']
            p['link'] = Connector.convert_path(href)

            # get article title
            for title in patch.xpath("./*[contains(@class, 'title-holder')]/*/text()"):
                p['title'] = title.strip()
                m = re.search(self.__regex_website_version, title)
                if m:
                    p['version'] = m.group(1)

            # get background image
            for style in patch.xpath("./*[contains(@class, 'background')]/@style"):
                m = re.match(self.__regex_website_background, style)
                if m:
                    background = m.group(2)
                    p['image'] = Connector.convert_path(background)

            result.append(p)
            print(p)
        return result



"""Starmap
"""
from .connector import Connector
from .interface import ICommand


class StarmapSystems(ICommand):
    """StarmapSystems
    """

    def __init__(self, name: str = None):
        """
        Args:
            name (str, optional): The name of the System. Default to None.
        """
        self.name = name

    async def execute_async(self):
        return self.execute()

    # Get Systems info
    def execute(self):
        systems = get_data("systems")
        if systems is None:
            return None

        if self.name is None:
            res = systems
        else:
            res = list(filter(lambda s: (str(s['name']).lower() == str(self.name).lower()), systems))

        if len(res) == 0:
            return None
        return res


class StarmapTunnels(ICommand):
    """StarmapTunnels
    """

    def __init__(self, tid: str = None):
        """

        Args:
            tid (str, optional): The tunnel identifier (tid). Default to None.
        """
        self.tid = tid

    async def execute_async(self):
        return self.execute()

    # Get Tunnels info
    def execute(self):
        tunnels = get_data("tunnels")
        if tunnels is None:
            return None

        if self.tid is None:
            res = tunnels
        else:
            res = list(filter(lambda s: str(s['id']).lower() == str(self.tid).lower(), tunnels))

        if len(res) == 0:
            return None
        return res


class StarmapSpecies(ICommand):
    """StarmapSpecies
    """

    def __init__(self, name: str = None):
        """
        Args:
            name (str, optional): The name of the specie. Default to None.
        """
        self.name = name

    async def execute_async(self):
        return self.execute()

    # Get Species info
    def execute(self):
        species = get_data("species")
        if species is None:
            return None

        if self.name is None:
            res = species
        else:
            res = list(filter(
                lambda s: str(s['name']).lower() == str(self.name).lower() or str(s['code']).lower() == str(
                    self.name).lower(), species))

        if len(res) == 0:
            return None
        return res


class StarmapAffiliations(ICommand):
    """StarmapAffiliations
    """

    def __init__(self, name: str = None):
        """
        Args:
            name (str, optional): The name of the affiliation. Default to None.
        """
        self.name = name

    async def execute_async(self):
        return self.execute()

    # Get Affiliations info
    def execute(self):
        affiliations = get_data("affiliations")
        if affiliations is None:
            return None

        if self.name is None:
            res = affiliations
        else:
            res = list(filter(lambda s: s['name'] == self.name or s['code'] == self.name, affiliations))

        if len(res) == 0:
            return None
        return res


class StarmapStarSystems(ICommand):
    """StarmapStarSystems
    """

    __url_star_system = "https://robertsspaceindustries.com/api/starmap/star-systems/{0}"

    def __init__(self, code: str):
        """
        Args:
            code (str): The code of the star system.
        """
        self.code = code

    async def execute_async(self):
        return self.execute()

    # Get Celestial Objects info
    def execute(self):
        req = Connector().request(self.__url_star_system.format(self.code))
        if req is None or req.status_code != 200:
            return None

        resp = req.json()

        # check response
        if resp['success'] != 1 or "data" not in resp:
            return None

        return resp['data']['resultset']


class StarmapCelestialObjects(ICommand):
    """StarmapCelestialObjects
    """

    __url_celestial = "https://robertsspaceindustries.com/api/starmap/celestial-objects/{0}"

    def __init__(self, code: str):
        """
        Args:
            code (str): The code field of the object.
        """
        self.code = code

    async def execute_async(self):
        return self.execute()

    # Get Celetial Objects info
    def execute(self):
        req = Connector().request(self.__url_celestial.format(self.code))
        if req is None or req.status_code != 200:
            return None

        resp = req.json()

        # check response
        if resp['success'] != 1 or "data" not in resp:
            return None

        return resp['data']['resultset']


class StarmapSearch(ICommand):
    """StarmapSearch
    """

    __url_find = "https://robertsspaceindustries.com/api/starmap/find"

    def __init__(self, search: str):
        """
        Args:
            search (str): The search.
        """
        self.search = search

    async def execute_async(self):
        return self.execute()

    def execute(self):
        data = {
            "query": self.search
        }
        req = Connector().request(self.__url_find, data)
        if req is None or req.status_code != 200:
            return None

        resp = req.json()

        # check response
        if resp['success'] != 1 or "data" not in resp:
            return None

        res = {
            "systems": resp['data']['systems']['resultset'],
            "objects": resp['data']['objects']['resultset']
        }
        return res


class StarmapRouteSearch(ICommand):
    """StarmapRouteSearch
    """

    __url_find = "https://robertsspaceindustries.com/api/starmap/routes/find"

    def __init__(self, _from: str, to: str, ship_size: str):
        """
        Args:
            _from (str): The object code for the departing system
            to (str): The object code for the destination system
            ship_size (str): The size of the ship traveling. Valid values are S, M, L
        """
        self._from = _from
        self._to = to
        self._size = ship_size

    async def execute_async(self):
        return self.execute()

    def execute(self):
        data = {
            "departure": self._from,
            "destination": self._to,
            "ship_size": self._size
        }
        req = Connector().request(self.__url_find, data)
        if req is None or req.status_code != 200:
            return None

        resp = req.json()

        # check response
        if resp['success'] != 1 or "data" not in resp:
            return None

        res = {
            "shortest": resp['data']['shortest'],
            "leastjumps": resp['data']['leastjumps']
        }
        return res


__url_systems = "https://robertsspaceindustries.com/api/starmap/bootup"


def get_data(field: str):
    """Retrieved data from the website

    Args:
        field (str): The field to return

    Returns:
        dict: The selected data
    """
    req = Connector().request(__url_systems)
    if req is None or req.status_code != 200:
        return None

    resp = req.json()

    # check response
    if resp['success'] != 1 or "data" not in resp:
        return None

    return resp['data'][field]['resultset']



import json
from os.path import join, dirname
from jsonschema import validate


def assert_valid_schema(data, schema_file):
    """ Checks whether the given data matches the schema """

    schema = _load_json_schema(schema_file)
    return validate(data, schema)


def _load_json_schema(filename):
    """ Loads the given schema file """

    relative_path = join('schemas', filename)
    absolute_path = join(dirname(__file__), relative_path)

    with open(absolute_path) as schema_file:
        return json.loads(schema_file.read())



"""Telemetry
"""

from .connector import Connector
from .interface import ICommand


class Telemetry(ICommand):
    """Telemetry
    """

    __url_telemetry = "https://robertsspaceindustries.com/api/telemetry/v2/community/"

    def __init__(self, timetable: str, version: str):
        """
        Args:
            timetable (str): "DAY", "WEEK", "MONTH".
            version (str): The Star Citizen game version.
        """
        self.timetable = timetable
        self.version = version

    async def execute_async(self):
        return self.execute()

    def execute(self):
        query = f"?&timetable={self.timetable}&branch=sc-alpha-{self.version}"

        req = Connector().request(self.__url_telemetry + query, method="get")

        if req is None or req.status_code != 200:
            return None

        resp = req.json()

        # check response
        if resp['success'] != 1 or "data" not in resp:
            return None

        return resp['data']



"""Roadmap
"""
from .connector import Connector
from .interface import ICommand


class Roadmap(ICommand):
    """Roadmap
    """

    __url_roadmap_sc = "https://robertsspaceindustries.com/api/roadmap/v1/boards/1"
    __url_roadmap_s42 = "https://robertsspaceindustries.com/api/roadmap/v1/boards/2"

    def __init__(self, board: str, version: str = None):
        """
        Args:
            board (str): The roadmap board ("starcitizen" or "squadron42").
            version (str, optional): The starcitizen version. Default to None.
        """
        self.board = board
        self.version = version

    async def execute_async(self):
        return self.execute()

    def execute(self):
        req = None
        if self.board == "starcitizen":
            req = Connector().request(self.__url_roadmap_sc, method="get")
        elif self.board == "squadron42":
            req = Connector().request(self.__url_roadmap_s42, method="get")
        else:
            return None

        if req is None or req.status_code != 200:
            return None

        try:
            resp = req.json()
        except Exception as e:
            print(e, flush=True)
            return None

        # get html contents
        if resp['success'] != 1 or "data" not in resp or "releases" not in resp['data']:
            return None

        releases = resp['data']['releases']
        categories = resp['data']['categories']

        cats = {}
        for c in categories:
            cats[int(c['id'])] = c['name']

        for r in releases:
            for c in r['cards']:
                c['category_name'] = cats[int(c['category_id'])]

        if self.version is None:
            return releases

        for r in releases:
            if r['name'] == self.version:
                return [r]
        return []



"""Stats
"""
import re

from .connector import Connector
from .interface import ICommand


class Stats(ICommand):
    __url_crow_stat = "https://robertsspaceindustries.com/api/stats/getCrowdfundStats"
    __url_roadmap = "https://robertsspaceindustries.com/api/roadmap/v1/init"
    __regex_api_live = r'(live\s*version\s*:\s*(?P<live_api>(\d+(\.\d+)*))?)'
    __regex_api_ptu = r'(ptu\s*version\s*:\s*(?P<ptu_version>(\d+(\.\d+)*))?)'
    __regex_api_etf = r'(ptu\s*version\s*:\s*(?P<etf_version>(\d+(\.\d+)*)(.*?)((ETF)|(Evocati))))'

    async def execute_async(self):
        return self.execute()

    def execute(self):
        """ Get general info
        """

        # Get stats info
        stats = self.__get_crow_stats()
        if stats is None:
            return None

        fleet = stats['fleet']
        if fleet is not None:
            fleet = int(fleet)

        req = Connector().request(self.__url_roadmap, method="get")
        if req is None or req.status_code != 200:
            return None

        text = req.text
        live_pu = self.__get_live_pu(text)
        live_ptu = self.__get_live_ptu(text)
        live_etf = self.__get_live_etf(text)

        return {
            "fans": int(stats['fans']),
            "funds": float(stats['funds']) / 100,
            "fleet": fleet,
            'current_live': live_pu,
            'current_ptu': live_ptu,
            'current_etf': live_etf,
        }

    def __get_live_pu(self, text):
        m = re.search(self.__regex_api_live, text, re.IGNORECASE)
        if m:
            return m.group(3)
        return None

    def __get_live_ptu(self, text):
        m = re.search(self.__regex_api_ptu, text, re.IGNORECASE)
        if m:
            return m.group(3)
        return None

    def __get_live_etf(self, text):
        m = re.search(self.__regex_api_etf, text, re.IGNORECASE)
        if m:
            return m.group(3)
        return None

    def __get_crow_stats(self, chart="day"):
        """ Get all Crow stats

            Returns:
                dict: assoc array of stats.
        """
        data = {
            "alpha_slots": True,
            "chart": chart,
            "fans": True,
            "fleet": True,
            "funds": True,
        }

        req = Connector().request(self.__url_crow_stat, method="post", json_data=data)

        if req is None or req.status_code != 200:
            return None

        resp = req.json()

        # check response
        if resp['success'] != 1 or "data" not in resp:
            return None

        return resp['data']



import re

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("rsi_scraper/__init__.py", encoding="utf8") as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)

setuptools.setup(
    name="rsi-scraper",
    version=version,
    author="Urbain Corentin",
    url="https://github.com/Dymerz/RSI-Scraper",
    description="Web scaper for RSI",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    project_urls={
        "Bug Tracker": "https://github.com/Dymerz/RSI-Scraper/issues",
        "Chat": "https://github.com/Dymerz/RSI-Scraper/discussions"
    },
    classifiers=[
        'Topic :: Software Development :: Libraries :: Python Modules',
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords='starcitizen robertsspaceindustries rsi scraper scraping',
    python_requires='>=3.7',
    install_requires=[
        'requests>=2.25.1',
        'lxml>=4.6.3',
    ]
)



import pytest
from .validator import assert_valid_schema


@pytest.mark.organization_members
@pytest.mark.parametrize('sid, page', [
    ('SIBYLLA', 1),
    ('PROTECTORA', 3),
    ('CORCOP', 1),
    ('IPIX', 3),
])
def test_request(sid, page):
    from rsi_scraper import OrganizationMembers

    object = OrganizationMembers(sid, page=page)
    data = object.execute()
    assert data is not None

    assert_valid_schema(data, 'organization_members.json')



import pytest
from .validator import assert_valid_schema


@pytest.mark.ship
def test_request():
    from rsi_scraper import Ship

    object = Ship(name='Cutlass Black')
    data = object.execute()
    assert data is not None

    assert_valid_schema(data, 'ship.json')



"""Get ships
"""
import re
import asyncio
from lxml import html

from .connector import Connector
from .interface import ICommand


class Ship(ICommand):
    """Get ships
    """

    __url_search_ships = "https://robertsspaceindustries.com/api/store/getShips"
    __url_ships_id = "https://robertsspaceindustries.com/ship-matrix/index"

    def __init__(self, **kwargs):
        """Get ships
        Args:
            **kwargs: Arbitrary keyword arguments.

        Keyword Args:
            name (str):
                The name of the ship to search.
            classification (str):
                The type of ship to get, you can pass this parameter at multiple times. ( combat , transport , exploration , industrial , support , competition , ground , multi ).
            length_min (int):
                The minimal length (in meters).
            length_max (int):
                The maximal length (in meters).
            crew_min (int):
                The minimal crew count.
            crew_max (int):
                The maximal crew count.
            price_min (int):
                The minimal price (in US dollars).
            price_max (int):
                The maximal price (in US dollars).
            mass_min (int):
                The minimal mass (Kg).
            mass_max (int):
                The maximal mass (Kg).
            page_max (int):
                The numbers of pages to process (1 page contains 10 ships), one page takes around 15s to process.
            page (int):
                The page to fetch (starts at 1).
        """
        self.kwargs = kwargs

    async def execute_async(self):
        return asyncio.run(self.get_ships_pages_async())

    def execute(self):
        res = asyncio.run(self.get_ships_pages_async())
        return res

    async def get_ships_pages_async(self):
        """Get all ships using advanced search through pages.

        Returns:
            list: Ships.
        """
        name = self.convert_val(self.kwargs.get("name"))
        classification = self.convert_val(self.kwargs.get("classification"))
        length_min = self.convert_val(self.kwargs.get("length_min"))
        length_max = self.convert_val(self.kwargs.get("length_max"))
        crew_min = self.convert_val(self.kwargs.get("crew_min"))
        crew_max = self.convert_val(self.kwargs.get("crew_max"))
        price_min = self.convert_val(self.kwargs.get("price_min"))
        price_max = self.convert_val(self.kwargs.get("price_max"))
        mass_min = self.convert_val(self.kwargs.get("mass_min"))
        mass_max = self.convert_val(self.kwargs.get("mass_max"))
        page = self.convert_val(self.kwargs.get("page"))
        page_max = self.convert_val(self.kwargs.get("page_max"))

        if re.match(r"^\d+?$", str(page)):
            page = int(page)
        else:
            page = 1

        if re.match(r"^\d+?$", page_max):
            page_max = int(page_max)
        else:
            page_max = 1

        data = {
            'classification': classification,
            'itemType': 'ships',
            'length': self.http_formatter(length_min, length_max),
            'manufacturer_id': [],
            'mass': self.http_formatter(mass_min, mass_max),
            'max_crew': self.http_formatter(crew_min, crew_max),
            'msrp': self.http_formatter(price_min, price_max),
            'search': name,
            'page': page,
            'sort': 'id',
            'storefront': 'pledge',
            'type': '',
        }

        req = Connector().request(self.__url_search_ships, data)

        if req is None or req.status_code == 404:
            return {}
        elif req.status_code != 200:
            return None

        resp = req.json()

        # get html contents
        if resp['success'] != 1:
            return None

        ships = []

        tree = html.fromstring(resp['data']['html'])
        tasks_ship = []
        ships_res = []
        for v in tree.xpath("//*[contains(@class, 'ship-item')]/@data-ship-id"):
            # Create async task
            tasks_ship.append(asyncio.create_task(self.get_ships_async(v.strip())))

        # Wait tasks to finish
        for t in tasks_ship:
            await t
            result = t.result()
            if result is not None:
                ships_res.append(result)

        for resp_ship, req_page in ships_res:
            page_tree = html.fromstring(req_page.content)

            # Get the Ship price
            price = 0
            for price in page_tree.xpath("//*[@class='final-price']/@data-value"):
                price = int(str(price))
                p = int(str(price)) / 100

                if price == 0 or price > p:
                    price = p
            resp_ship['price'] = price

            ships.append(resp_ship)

        if resp['data']['rowcount'] != 0 and page < page_max:
            self.kwargs['page'] = page + 1
            val = self.execute(**self.kwargs)
            if val != []:
                ships.extend(val)
        return ships

    async def get_ships_async(self, ship_id: int):
        """Get ship by its id.

        Args:
            ship_id (int): The ship id.

        Returns:
            dict: The ship.
            requests.models.Response:
        """
        resp_ship = await self.get_ship_by_id(ship_id)
        req_page = None

        if resp_ship is not None:
            req_page = (await asyncio.gather(Connector().request_async(Connector().url_host + resp_ship['url'])))[0]

        if req_page is None or req_page.status_code != 200:
            return None
        return (resp_ship, req_page)

    async def get_ship_by_id(self, ship_id: int = None, get_price=False):
        """Get a ship by his ship_id.

        Args:
            ship_id (int, optional): The ship ID. Default to None.
            get_price (bool, optional): if the price need to be retrieved. Default to False.

        Returns:
            dict: The ship.
        """

        parameters = {}
        if ship_id is not None:
            parameters = {'id': ship_id}

        req = Connector().request(self.__url_ships_id, parameters)

        if req is None:
            return None
        elif req.status_code == 404:
            return {}
        elif req.status_code != 200:
            return None

        resp_ship = req.json()

        if resp_ship['success'] != 1 or 'data' not in resp_ship or len(resp_ship['data']) == 0:
            return None

        res = []
        if ship_id is None:
            res = resp_ship['data']
        else:
            res = resp_ship['data'][0]

        if get_price:
            for s in res:
                req_page = Connector().request(Connector().url_host + s['url'])

                page_tree = html.fromstring(req_page.content)

                # Get the Ship price
                price = 0
                for v in page_tree.xpath("//*[@class='final-price']/@data-value"):
                    p = int(str(v)) / 100

                    if price == 0 or price > p:
                        price = p
                s['price'] = price

        return res



import pytest
from .validator import assert_valid_schema


@pytest.mark.starmap
def test_request_systems():
    from rsi_scraper import StarmapSystems

    object = StarmapSystems('pyro')
    data = object.execute()
    assert data is not None

    assert_valid_schema(data, 'starmap_systems.json')


@pytest.mark.starmap
def test_request_search():
    from rsi_scraper import StarmapSearch

    object = StarmapSearch('Stanton')
    data = object.execute()
    assert data is not None

    assert_valid_schema(data, 'starmap_search.json')


@pytest.mark.starmap
def test_request_object():
    from rsi_scraper import StarmapCelestialObjects

    object = StarmapCelestialObjects('TOHIL.PLANETS.TOHILIII')
    data = object.execute()
    assert data is not None

    assert_valid_schema(data, 'starmap_object.json')


@pytest.mark.starmap
def test_request_star_system():
    from rsi_scraper import StarmapStarSystems

    object = StarmapStarSystems('STANTON')
    data = object.execute()
    assert data is not None

    assert_valid_schema(data, 'starmap_star_system.json')


@pytest.mark.starmap
def test_request_tunnels():
    from rsi_scraper import StarmapTunnels

    object = StarmapTunnels('1188')
    data = object.execute()
    assert data is not None

    assert_valid_schema(data, 'starmap_tunnels.json')


@pytest.mark.starmap
def test_request_species():
    from rsi_scraper import StarmapSpecies

    object = StarmapSpecies()
    data = object.execute()
    assert data is not None

    assert_valid_schema(data, 'starmap_species.json')


@pytest.mark.starmap
@pytest.mark.parametrize('_from, to, ship_size', [
    ('STANTON.LZS.LORVILLE', 'STANTON.LZS.AREA18', "M"),
    ('STANTON.LZS.LORVILLE', 'STANTON.LZS.AREA18', "L"),
    ('STANTON.LZS.LORVILLE', 'STANTON.LZS.AREA18', "S"),
    ('SOL.PLANETS.EARTH', 'PYRO.PLANETS.PYROI', 'L'),
    ('ORION.PLANET.ORIONIIIARMITAGE', 'AYRKA.PLANET.AYRKAII', 'S'),
])
def test_request_route_search(_from, to, ship_size):
    from rsi_scraper import StarmapRouteSearch

    object = StarmapRouteSearch(_from, to, ship_size)
    data = object.execute()

    assert data is not None

    assert_valid_schema(data, 'starmap_route_search.json')



import pytest
from .validator import assert_valid_schema


@pytest.mark.version
def test_request():
    from rsi_scraper import Version

    object = Version()
    data = object.execute()
    assert data is not None

    assert_valid_schema(data, 'version.json')



"""Get User
"""
from datetime import datetime
from lxml import html

from .connector import Connector
from .interface import ICommand


class User(ICommand):
    """Get User
    """

    __url_profile = "https://robertsspaceindustries.com/citizens/{0}"
    __url_affiliation = "https://robertsspaceindustries.com/citizens/{0}/organizations"

    def __init__(self, user_handle):
        """Get user info.

        Args:
            user_handle (str): The user handle.
        """
        super().__init__()
        self.user_handle = user_handle

    async def execute_async(self):
        return self.execute()

    def execute(self):
        return self.get_user_info(self.user_handle)

    def get_user_info(self, user_handle: str):
        """Get the user profile.

        Args:
            user_handle (str): The user handle.

        Returns:
            dict: The user info.
        """

        # format url
        url = self.__url_profile.format(user_handle)

        # request website
        req = Connector().request(url, method="get")

        if req is None or req.status_code == 404:
            return {}
        elif req.status_code != 200:
            return None

        # get html contents
        tree = html.fromstring(req.content)

        # begin store basic data
        result = {"profile": {}, "organization": {}}

        result["profile"]["page"] = {}
        result["profile"]["page"]["url"] = url

        # get identifier
        for v in tree.xpath('//title/text()'):
            result["profile"]["page"]["title"] = v.strip()
            break

        # get identifier
        for v in tree.xpath('//*[contains(@class, "label") and text() = "UEE Citizen Record"]/following-sibling::*[1]/text()'):
            result["profile"]["id"] = v.strip()
            break

        # get displayed name
        for v in tree.xpath('//*[@class="info"][1]/p[1]/*[@class="value"]/text()'):
            result["profile"]["display"] = v.strip()
            break

        # get handle
        for v in tree.xpath('//*[contains(@class, "label") and text() = "Handle name"]/following-sibling::*[1]/text()'):
            result["profile"]["handle"] = v.strip()
            break

        # get badge
        for v in tree.xpath('//*[@class="info"][1]/p[last()]/*[@class="value"]/text()'):
            result["profile"]["badge"] = v.strip()
            break

        # get badge image
        for v in tree.xpath('//*[@class="info"][1]/*[@class="entry"]/*[@class="icon"]/img/@src'):
            d = v.strip()
            if d[0:1] == '/':
                d = "https://robertsspaceindustries.com" + d
            result["profile"]["badge_image"] = d
            break

        # get profile image
        for v in tree.xpath('//*[contains(@class, "title") and contains(text(), "Profile")]/following-sibling::*/div[@class="thumb"]/img/@src'):
            result["profile"]["image"] = Connector.convert_path(v.strip())
            break

        # get organization image
        for v in tree.xpath('//*[contains(@class, "title") and contains(text(), "Main organization")]/following-sibling::*/div[@class="thumb"]/a/img/@src'):
            result["organization"]["image"] = Connector.convert_path(v.strip())
            break

        # get organization name
        for v in tree.xpath('//a[contains(@class, "value") and contains(@class, "data")]/text()'):
            result["organization"]["name"] = v.strip()
            break

        # get organization SID (Spectrum Identification)
        for v in tree.xpath('//*[contains(@class, "label") and text() = "Spectrum Identification (SID)"]/following-sibling::*[1]/text()'):
            result["organization"]["sid"] = v.strip()
            break

        # get organization rank
        for v in tree.xpath('//*[contains(@class, "label") and text() = "Organization rank"]/following-sibling::*[1]/text()'):
            result["organization"]["rank"] = v.strip()
            break

        # get organization stars
        result["organization"]["stars"] = len(tree.xpath(".//*[contains(@class, 'ranking')]/span[contains(@class, 'active')]"))

        # get enlist date
        for v in tree.xpath('//*[contains(@class, "label") and text() = "Enlisted"]/following-sibling::*[1]/text()'):
            result["profile"]["enlisted"] = datetime.strptime(v.strip(), "%b %d, %Y").strftime('%Y-%m-%dT%H:%M:%S.%f')
            break

        # find country and/or region
        for v in tree.xpath('//*[contains(@class, "label") and text() = "Location"]/following-sibling::*[@class="value"]/text()'):
            result["profile"]["location"] = {}
            if "," in v.strip():
                arr = v.split(",")  # location splitter
                result["profile"]["location"]["country"] = arr[0].strip()
                result["profile"]["location"]["region"] = arr[1].strip()
            else:
                result["profile"]["location"]["country"] = v.strip()
            break

        # get fluency
        for v in tree.xpath('//*[contains(@class, "label") and text() = "Fluency"]/following-sibling::*[1]/text()'):
            result["profile"]["fluency"] = [x.strip() for x in v.split(",")]
            break

        # get website
        for v in tree.xpath('//*[contains(@class, "label") and text() = "Website"]/following-sibling::*/text()'):
            result["profile"]["website"] = v.strip()
            break

        # get bio
        result["profile"]["bio"] = ""
        for v in tree.xpath('//*[contains(@class, "label") and text() = "Bio"]/following-sibling::*[@class="value"]//text()'):
            result["profile"]["bio"] += v
        result["profile"]["bio"] = result["profile"]["bio"].strip()

        # remove useless bio
        if result["profile"]["bio"] == "":
            del result["profile"]["bio"]

        result["affiliation"] = self._get_user_affiliations(user_handle)

        return result

    def _get_user_affiliations(self, user_handle: str):
        """Get user affiliations.

        Args:
            user_handle (str): The user handle.

        Returns:
            dict: The user affiliations.
        """

        # format url
        url = self.__url_affiliation.format(user_handle)

        # request website
        req = Connector().request(url, method="get")

        if req is None or req.status_code == 404:
            return {}
        elif req.status_code != 200:
            return None

        # get html contents
        tree = html.fromstring(req.content)

        # store basic data
        result = []

        for a in tree.xpath("//*[contains(@class, 'affiliation')]"):
            aff = {}

            for v in a.xpath(".//*[contains(@class, 'entry')]/strong[contains(../span, 'SID')]/text()"):
                aff["sid"] = v.strip()
                break

            for v in a.xpath(".//*[contains(@class, 'orgtitle')]/a/text()"):
                aff["name"] = v.strip()
                break

            for v in a.xpath(".//*[contains(@class, 'entry')]/strong[contains(../span, 'rank')]/text()"):
                aff["rank"] = v.strip()
                break
            aff["stars"] = len(a.xpath(".//*[contains(@class, 'ranking')]/span[contains(@class, 'active')]"))

            for v in a.xpath(".//img/@src"):
                aff["image"] = Connector.convert_path(v.strip())
                break

            result.append(aff)

        return result
